/*
 *  jDTAUS - DTAUS fileformat.
 *  Copyright (c) 2005 Christian Schulte <cs@schulte.it>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
package org.jdtaus.core.monitor.ri.client;

import java.util.HashMap;
import java.util.Map;
import javax.swing.event.EventListenerList;
import org.jdtaus.core.container.ContainerInitializer;
import org.jdtaus.core.container.Dependency;
import org.jdtaus.core.container.Implementation;
import org.jdtaus.core.container.ImplementationException;
import org.jdtaus.core.container.ModelFactory;
import org.jdtaus.core.container.Property;
import org.jdtaus.core.container.PropertyException;
import org.jdtaus.core.monitor.Task;
import org.jdtaus.core.monitor.TaskEvent;
import org.jdtaus.core.monitor.TaskListener;
import org.jdtaus.core.monitor.spi.TaskMonitor;

/**
 * jDTAUS Core SPI Client {@code TaskMonitor} implementation.
 * <p>This client implementation will start a new monitoring thread when
 * requested with the {@code monitor()} method polling the state of the task
 * every {@code pollIntervalMillis} milliseconds to fire corresponding
 * {@code TaskEvent}s.</p>
 * <p><b>Note:</b><br/>
 * {@code TaskEvent}s produced by this implementation will be fired from the
 * monitor thread, not from the thread executing the task.</p>
 *
 * @author <a href="mailto:cs@schulte.it">Christian Schulte</a>
 * @version $Id: ClientTaskMonitor.java 2526 2007-04-12 09:47:08Z schulte2005 $
 *
 * @see org.jdtaus.core.container.Container
 */
public class ClientTaskMonitor implements TaskMonitor, ContainerInitializer
{

    //--Implementation----------------------------------------------------------

    // This section is generated by jdtaus-source-plugin.

    /** Meta-data describing the implementation. */
    private static final Implementation META =
        ModelFactory.getModel().getModules().
        getImplementation(ClientTaskMonitor.class.getName());

    //----------------------------------------------------------Implementation--
    //--Constructors------------------------------------------------------------

    // This section is generated by jdtaus-source-plugin.

    /** Protected <code>ClientTaskMonitor</code> implementation constructor.
    * @param meta Implementation meta-data.
    */
    protected ClientTaskMonitor(final Implementation meta)
    {
        super();
        Property p;

        p = meta.getProperties().getProperty("pollIntervalMillis");
        this._pollIntervalMillis = ((java.lang.Long) p.getValue()).longValue();

    }
    /** Protected <code>ClientTaskMonitor</code> dependency constructor.
    * @param meta Dependency meta-data.
    */
    protected ClientTaskMonitor(final Dependency meta)
    {
        super();
        Property p;

        p = meta.getProperties().getProperty("pollIntervalMillis");
        this._pollIntervalMillis = ((java.lang.Long) p.getValue()).longValue();

    }

    //------------------------------------------------------------Constructors--
    //--ContainerInitializer----------------------------------------------------

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for invalid property values.
     *
     * @see #assertValidProperties()
     */
    public void initialize()
    {
        this.assertValidProperties();
    }

    //----------------------------------------------------ContainerInitializer--
    //--Properties--------------------------------------------------------------

    // This section is generated by jdtaus-source-plugin.

    /**
     * Property {@code pollIntervalMillis}.
     * @serial
     */
    private long _pollIntervalMillis;

    /** <code>pollIntervalMillis</code> property getter. */
    protected long getPollIntervalMillis()
    {
        return this._pollIntervalMillis;
    }


    //--------------------------------------------------------------Properties--
    //--TaskEventSource---------------------------------------------------------

    /** List of {@code TaskListener}s. */
    private final EventListenerList taskListeners = new EventListenerList();

    public final void addTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        synchronized(this.taskListeners)
        {
            this.taskListeners.add(TaskListener.class, listener);
        }
    }

    public final void removeTaskListener(final TaskListener listener)
    {
        if(listener == null)
        {
            throw new NullPointerException("listener");
        }

        synchronized(this.taskListeners)
        {
            this.taskListeners.remove(TaskListener.class, listener);
        }
    }

    public final TaskListener[] getTaskListeners()
    {
        synchronized(this.taskListeners)
        {
            return (TaskListener[]) this.taskListeners.
                getListeners(TaskListener.class);

        }
    }

    /**
     * Notifies all registered {@code TaskListener}s about
     * {@code TaskEvent}s.
     *
     * @param e The event to be provided to the listeners.
     */
    protected final void fireTaskEvent(final TaskEvent e)
    {
        if(e == null)
        {
            throw new NullPointerException("e");
        }

        synchronized(this.taskListeners)
        {
            final Object[] listeners = this.taskListeners.getListenerList();
            for (int i = listeners.length - 2; i >= 0; i -= 2)
            {
                if (listeners[i] == TaskListener.class)
                {
                    ((TaskListener) listeners[i + 1]).onTaskEvent(e);
                }
            }
        }
    }

    //---------------------------------------------------------TaskEventSource--
    //--TaskMonitor-------------------------------------------------------------

    /** Maps {@code Task}s to theire monitoring threads. */
    private final Map monitorMap = new HashMap(100);

    public void monitor(final Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        final MonitorThread t;
        synchronized(this.monitorMap)
        {
            this.finish(task);
            t = new MonitorThread(task, this.getPollIntervalMillis());
            this.monitorMap.put(task, t);
            t.start();
        }
    }

    public void finish(Task task)
    {
        if(task == null)
        {
            throw new NullPointerException("task");
        }

        final MonitorThread t;
        synchronized(this.monitorMap)
        {
            if((t = (MonitorThread) this.monitorMap.remove(task)) != null)
            {
                try
                {
                    t.setRunning(false);
                    t.interrupt();
                    t.join();
                }
                catch(InterruptedException e)
                {
                    throw new ImplementationException(
                        ClientTaskMonitor.META, e);

                }
            }
        }
    }

    //-------------------------------------------------------------TaskMonitor--
    //--ClientTaskMonitor-------------------------------------------------------

    /** Creates a new {@code ClientTaskMonitor} instance. */
    public ClientTaskMonitor()
    {
        this(ClientTaskMonitor.META);
        this.initialize();
    }

    /** Thread monitoring a {@code Task}. */
    public class MonitorThread extends Thread
    {

        private boolean running;
        private final long pollIntervalMillis;
        private final Task task;

        /**
         * Creates a new {@code MonitorThread} instance.
         *
         * @param task the task to monitor.
         * @param pollIntervalMillis amount of milliseconds to wait before
         * checking the state of {@code task} for changes.
         *
         * @throws NullPointerException if {@code task} is {@code null}.
         * @throws IllegalArgumentException if {@code pollIntervalMillis} is
         * negative or equal to zero.
         */
        protected MonitorThread(final Task task,
            final long pollIntervalMillis)
        {

            if(task == null)
            {
                throw new NullPointerException("task");
            }
            if(pollIntervalMillis <= 0L)
            {
                throw new IllegalArgumentException(
                    Long.toString(pollIntervalMillis));

            }

            this.task = task;
            this.running = true;
            this.pollIntervalMillis = pollIntervalMillis;
        }

        /** {@inheritDoc} */
        public void run()
        {
            boolean indeterminate = this.task.isIndeterminate();
            int minimum = this.task.getMinimum();
            int maximum = this.task.getMaximum();
            int progress = this.task.getProgress();
            fireTaskEvent(new TaskEvent(this.task, TaskEvent.STARTED));
            while(this.isRunning())
            {
                try
                {
                    this.sleep(this.pollIntervalMillis);
                    if(this.task.getMinimum() != minimum ||
                        this.task.getMaximum() != maximum ||
                        this.task.getProgress() != progress ||
                        this.task.isIndeterminate() != indeterminate)
                    {

                        fireTaskEvent(new TaskEvent(this.task,
                            TaskEvent.CHANGED_STATE));

                        indeterminate = this.task.isIndeterminate();
                        minimum = this.task.getMinimum();
                        maximum = this.task.getMaximum();
                        progress = this.task.getProgress();
                    }
                }
                catch(InterruptedException e)
                {
                    this.setRunning(false);
                }
            }
            fireTaskEvent(new TaskEvent(this.task, TaskEvent.ENDED));
        }

        /**
         * Getter for property {@code running}.
         *
         * @return {@code true} if the thread is currently running;
         * {@code false} if not.
         */
        public boolean isRunning()
        {
            return this.running;
        }

        /**
         * Setter for property {@code running}.
         * <p/><b>Note:</b><br/>
         * To stop a thread the instance needs to be interrupted after setting
         * this property to {@code false}.
         *
         * @param running {@code true} to mark the thread as currently running;
         * {@code false} to stop the thread.
         */
        public void setRunning(final boolean running)
        {
            this.running = running;
        }
    }

    /**
     * Checks configured properties.
     *
     * @throws PropertyException for illegal property values.
     */
    protected void assertValidProperties()
    {
        if(this.getPollIntervalMillis() < 100L)
        {
            throw new PropertyException("pollIntervalMillies",
                Long.toString(this.getPollIntervalMillis()));

        }
    }

    //-------------------------------------------------------ClientTaskMonitor--

}
